from collections import OrderedDict

import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator
import matplotlib.image as image
from munch import *

# Default papersizes in inches and choose one according to dict
PAPER_SIZES = dict(A3=[16.5, 11.75], A4=[8.25, 11.75],)  # A3 Landscape  # A4 Portrait


def _fig_to_ax(fig, ax, coords):
    display_coords = fig.transFigure.transform(coords)
    inv = ax.transAxes.inverted()
    return inv.transform(display_coords)


def _ax_to_fig(fig, ax, coords):
    display_coords = ax.transAxes.transform(coords)
    inv = fig.transFigure.inverted()
    return inv.transform(display_coords)


def cm2inch(cm):
    inch = 2.54
    if isinstance(cm, tuple):
        return tuple(i/inch for i in cm)
    else:
        return cm/inch


def plot_si_results(plot_props):
    """Function to plot si results in an gINT like manner

    Parameters
    ----------
    plot_props : dict
        Dictionary containing all data and kwargs for the plots
    """
    # if plot props is not a munch, turn it into one.
    if not isinstance(plot_props, Munch):
        plot_props = munchify(plot_props)

    # cm to inches
    cm_to_inch = 1/2.54

    papersize = PAPER_SIZES[plot_props.layout.papersize]

    # Change some rcParams
    plt.rcParams["font.family"] = "arial"
    plt.rcParams["font.size"] = 8.5
    plt.rcParams["figure.dpi"] = 80
    plt.rcParams["savefig.dpi"] = 400
    plt.rcParams['hatch.linewidth'] = 0.1

    # initialize figure, size in inches
    fig = plt.figure(figsize=papersize)

    # set margin around figure
    margins = Munch(
        left=(cm_to_inch * plot_props.layout.margins[0]),
        right=(cm_to_inch * plot_props.layout.margins[1]),
        top=(cm_to_inch * plot_props.layout.margins[2]),
        bottom=(cm_to_inch * plot_props.layout.margins[3]),
    )

    fig_layout = Munch()
    fig_layout.dimensions = Munch()

    # main dimensions axis in cm, w x h. List comprehension to convert to inches
    fig_layout.dimensions.h_header = 3.5 * cm_to_inch
    fig_layout.dimensions.h_spacing_header_title = 0.1 * cm_to_inch
    fig_layout.dimensions.h_title = 1.0 * cm_to_inch
    fig_layout.dimensions.h_spacing_axis_footer = 0.1 * cm_to_inch
    fig_layout.dimensions.h_footer = 2.0 * cm_to_inch
    fig_layout.dimensions.level_box_indent = 1.25 * cm_to_inch
    fig_layout.dimensions.logo = 1.5 * cm_to_inch

    fig_layout.dimensions.h_axis = (
        papersize[1]
        - margins.top
        - margins.bottom
        - sum(fig_layout.dimensions[h] for h in fig_layout.dimensions if 'h_' in h)
    )

    fig_layout.dimensions.w_axis = (
        papersize[0]
        - margins.left
        - margins.right
        - fig_layout.dimensions.level_box_indent
    )
    fig_layout.dimensions.h_level = fig_layout.dimensions.h_axis

    fig_layout.axes = Munch()
    fig_layout.axes.dimensions = Munch()

    for i in range(plot_props.axes.num):
        fig_layout.axes.dimensions["ax" + str(i + 1)] = [
            plot_props.axes.widths[i] * fig_layout.dimensions.w_axis,
            fig_layout.dimensions.h_axis,
        ]

    fig_layout.axes.dimensions.header = [
        fig_layout.dimensions.w_axis + fig_layout.dimensions.level_box_indent,
        fig_layout.dimensions.h_header,
    ]
    fig_layout.axes.dimensions.title = [
        fig_layout.dimensions.w_axis + fig_layout.dimensions.level_box_indent,
        fig_layout.dimensions.h_title,
    ]
    fig_layout.axes.dimensions.level = [
        fig_layout.dimensions.level_box_indent,
        fig_layout.dimensions.h_axis,
    ]
    fig_layout.axes.dimensions.footer = [
        fig_layout.dimensions.w_axis + fig_layout.dimensions.level_box_indent,
        fig_layout.dimensions.h_footer,
    ]

    fig_layout.axes.rect = Munch()

    # make header boxes
    yll_offset = 0
    for box in [
        "footer",
        "h_spacing_axis_footer",
        "level",
        "title",
        "h_spacing_header_title",
        "header",
    ]:
        if "spacing" not in box:
            fig_layout.axes.rect[box] = [
                margins.left / papersize[0],  # xll
                (margins.bottom + yll_offset) / papersize[1],  # yll
                fig_layout.axes.dimensions[box][0] / papersize[0],  # dx
                fig_layout.axes.dimensions[box][1] / papersize[1],  # dy
            ]

            yll_offset = yll_offset + fig_layout.dimensions["h_" + box]
        else:
            yll_offset = yll_offset + fig_layout.dimensions[box]

    # make axis for plotting
    xll_offset = fig_layout.dimensions.level_box_indent
    for i in range(plot_props.axes.num):
        fig_layout.axes.rect["ax" + str(i + 1)] = [
            (margins.left + xll_offset) / papersize[0],  # xll
            fig_layout.axes.rect.level[1],
            fig_layout.axes.dimensions["ax" + str(i + 1)][0] / papersize[0],
            fig_layout.axes.dimensions["ax" + str(i + 1)][1] / papersize[1],
        ]

        xll_offset = xll_offset + fig_layout.axes.dimensions["ax" + str(i + 1)][0]

        # plot the info headers
    fig_layout.axes.axs_info = Munch()
    for ax in ["footer", "level", "title", "header"]:
        fig_layout.axes.axs_info[ax] = plt.axes(
            fig_layout.axes.rect[ax], facecolor="none"
        )

    # Place logo..
    im_xy0 = _ax_to_fig(fig, fig_layout.axes.axs_info.header, [0.00, 0.65])
    ax_logo = plt.axes(
        [
            im_xy0[0] + (0.25 * cm_to_inch / papersize[1]),
            im_xy0[1] + (0.10 * cm_to_inch / papersize[1]),
            (3 * cm_to_inch / papersize[0]),
            (1 * cm_to_inch / papersize[1]),
        ]
    )
    ax_logo.axis("off")

    if 'logo' in plot_props.layout.keys():
        if plot_props.layout.logo is not None:
            ax_logo.imshow(
                image.imread(plot_props.layout.logo),
                zorder=-1,
                origin="upper",
                interpolation="nearest",
            )

    # plot the graphs axes..
    fig_layout.axes.axs_plots = Munch()
    for i in range(plot_props.axes.num):
        fig_layout.axes.axs_plots["ax" + str(i + 1)] = plt.axes(
            fig_layout.axes.rect["ax" + str(i + 1)],
            **plot_props.axes.kwargs["ax" + str(i + 1)],
            **plot_props.axes.kwargs.all
        )

    # make twiny axes if any..
    if "twiny" in [x for x in plot_props.axes]:
        for twiny in plot_props.axes.twiny:
            ref_ax = plot_props.axes.twiny[twiny].twin
            fig_layout.axes.axs_plots[twiny] = fig_layout.axes.axs_plots[ref_ax].twiny()

            ax = fig_layout.axes.axs_plots[twiny]
            kwargs = plot_props.axes.twiny[twiny].kwargs
            for key in kwargs:
                getattr(ax, "set_%s" % key)(kwargs.get(key))

    # ax formatting
    try:
        no_grid = plot_props.axes.no_grid
    except AttributeError:
        no_grid = []

    for key in fig_layout.axes.axs_plots:
        ax = fig_layout.axes.axs_plots[key]

        if key in no_grid:
            pass
        else:
            ax.grid(linestyle="--")

        ax.xaxis.set_label_position("top")
        ax.xaxis.tick_top()
        ax.yaxis.set_minor_locator(MultipleLocator(0.25))

        # set the alignment for outer ticklabels
        ticklabels = ax.get_xticklabels()
        ticklabels[0].set_ha("left")
        ticklabels[-1].set_ha("right")

        # set the alignment for outer ticklabels
        ticklabels = ax.get_yticklabels()
        ticklabels[-1].set_va("top")
        ticklabels[0].set_va("bottom")

        # move secondary x-axis to the inside
        if "_" in key:
            ax.tick_params(axis="x", direction="in", pad=-15)
        else:
            ax.tick_params(axis="both", pad=1)

    # turn of labels and/or ticks for some of the axis
    for key in fig_layout.axes.axs_plots:
        ax = fig_layout.axes.axs_plots[key]
        if "ax1" not in key:
            ax.set_yticklabels([])

    for key in fig_layout.axes.axs_info:
        ax = fig_layout.axes.axs_info[key]
        ax.set_xticks([])
        ax.set_yticks([])

    # place divider lines in footer and title block
    for i in range(plot_props.axes.num):
        y = fig_layout.axes.rect.title[1]
        x = fig_layout.axes.rect["ax" + str(i + 1)][0]
        x_t = _fig_to_ax(fig, fig_layout.axes.axs_info.title, [x, y])[0]
        for ax in ["footer", "title"]:
            fig_layout.axes.axs_info[ax].axvline(x_t, color="black", lw=0.75)

    # make the plots!!
    y_data = plot_props.plots.y_data

    for plot in plot_props.plots.plots:
        ax = fig_layout.axes.axs_plots[plot.ax]
        kwargs = plot.kwargs
        plot_type = plot.type

        if 'x_data' in plot.keys():
            args = (plot.x_data, y_data)
        elif 'args' in plot.keys():
            args = tuple(plot.args)
        else:
            raise ValueError('No x-data passed??')

        getattr(ax, plot_type)(*args, **kwargs)

    # legends
    for key in fig_layout.axes.axs_plots:

        # filter out the twin y axis to avoid legend mistakes.
        if "_" not in key:
            ax = fig_layout.axes.axs_plots[key]
            handles, labels = ax.get_legend_handles_labels()

            # get handles from twin y axis if available.
            try:
                ax_2 = fig_layout.axes.axs_plots[key + "_2"]
                handles_2, labels_2 = ax_2.get_legend_handles_labels()
                handles = handles + handles_2
                labels = labels + labels_2
            except KeyError:
                pass

            # Filter out double entries
            by_label = OrderedDict(zip(labels, handles))
            # Set position, place top left point of legend at bottom left point of ax.
            bbox = ax.get_position()
            xy = (bbox.xmin, bbox.ymin)
            legend_kwargs = dict(
                loc="upper left",
                bbox_to_anchor=xy,
                bbox_transform=fig.transFigure,
                frameon=False,)
            # Check if custom legend is available
            try:
                ax.legend(
                    handles=plot_props.axes.custom_legend[key],
                    fontsize='x-small',
                    ncol=1,
                    **legend_kwargs
                    )
            except (KeyError, AttributeError):
                ax.legend(
                    by_label.values(),
                    by_label.keys(),
                    fontsize='x-small',
                    ncol=3,
                    **legend_kwargs
                )

    # Header info and styling
    hline_y = 0.65
    fig_layout.axes.axs_info.header.axhline(hline_y, lw=0.80, color="black")

    header = fig_layout.axes.axs_info.header
    sep = ":"
    sep_cols = [1, 4, 7]
    cols = [key for key in plot_props.info if "titles" in key or "values" in key]
    cols_spacing = [
        0.0125,
        0.1300,
        0.1500,
        0.5000,
        0.6175,
        0.6375,
        0.7500,
        0.8675,
        0.8875,
    ]

    if plot_props.layout.papersize == "A3":
        cols_spacing = [s * 0.5 for s in cols_spacing]

    for s in sep_cols:
        cols.insert(s, sep)

    for i, x in enumerate(cols_spacing):
        for j, y in enumerate(plot_props.info.row_spacing):
            if cols[i] == sep:
                header.text(
                    x, y, sep, verticalalignment="top", horizontalalignment="left"
                )
            else:
                header.text(
                    x,
                    y,
                    plot_props.info[cols[i]][j],
                    verticalalignment="top",
                    horizontalalignment="left",
                )

    header.text(
        0.99,
        1.0 - ((1 - hline_y) / 2),
        " - ".join(plot_props.info.pointid),
        verticalalignment="center",
        horizontalalignment="right",
        fontsize=16,
        fontweight="bold",
    )

    plt.close()
    return fig



from typing import List, Optional, Union, Dict, Any
from typing_extensions import Literal

from pydantic import BaseModel, FilePath, Field, validator
from pydantic.color import Color
import numpy as np
from matplotlib.colors import ListedColormap, BoundaryNorm


def _convert_color(color: Color) -> str:
    if isinstance(color, Color):
        return color.as_hex()
    else:
        return color


class LineKwargs(BaseModel):
    label: str = Field(..., description="Label to appear in legend")
    color: Color = Field("black", description="Line color")
    ls: Literal[
        "-", "--", "-.", ":", " ", "", "solid", "dashed", "dashdot", "dotted"
    ] = Field("-", description="Line style of plot")
    lw: float = Field(1.5, lt=5.0, gt=0.1, description="Line width of plot")

    _normalize_color = validator("color", allow_reuse=True)(_convert_color)


class ScatterKwargs(BaseModel):
    label: str = Field(..., description="Label to appear in legend")
    c: Union[
        Color, List[Union[float, None]]
    ] = None  # Field("white", description="Face color of marker")
    cmap: Any = None
    norm: Any = None
    edgecolor: Color = None  # Field("black", description="Edge color of marker")
    s: int = Field(100, gt=1, lt=150, description="Marker size")
    marker: Literal["^", "v", "<", ">", "o", "x"] = None  # Field(
    #     "^", description="Marker style"
    # )

    _normalize_c = validator("c", allow_reuse=True)(_convert_color)
    _normalize_edgecolor = validator("edgecolor", allow_reuse=True)(_convert_color)


class BarKwargs(BaseModel):
    label: str = Field(..., description="Label to appear in legend")
    width: Optional[float] = Field(0.5, gt=0.1, lt=2, description="Width of bar")
    bottom: List[float] = Field(
        ..., description="List to indicate y position of bottom bar"
    )
    color: List[str] = Field(..., description="List of bar colors")
    hatch: Optional[Literal["...", "..", "\\", "/"]] = Field(
        None, description="Hatch style"
    )

    # _normalize_color = validator("color", allow_reuse=True)(_convert_color)


class AxKwargs(BaseModel):
    xlim: List[float] = None
    xlabel: str = None
    ylim: List[float] = None
    ylabel: str = None
    xticks: Optional[List[float]]
    yticks: Optional[List[float]]
    xticklabels: Optional[List[str]]
    yticklabels: Optional[List[str]]


class Plot(BaseModel):
    ax: str
    name: str
    type: Literal["plot", "scatter", "axvline", "axhline", "bar"] = "plot"
    x_data: List[float] = None
    args: List[List[Union[float, None]]] = None
    kwargs: Optional[Any]


class PlotContainer(BaseModel):
    y_data: List[float]
    plots: List[Plot]


class Twiny(BaseModel):
    twin: str
    kwargs: AxKwargs


class Axes(BaseModel):
    num: int
    widths: List[float]
    kwargs: Dict[str, AxKwargs]
    twiny: Optional[Dict[str, Twiny]]

    @validator("widths")
    def validate_widths(cls, v):
        assert sum(v) == 1.0, "sum of all widths must be 1"
        return v


class Layout(BaseModel):
    papersize: Literal["A4", "A3"] = "A4"
    margins: List[float] = [1, 1, 1, 1]
    logo: Optional[FilePath]


class Info(BaseModel):
    pointid: List[str]
    row_spacing: List[float] = np.linspace(0.6, 0.1, 5)
    general_titles: List[str] = ["CLIENT", "ENGINEER", "AC", "CONTRACTOR", "PROJECT"]
    general_values: List[str]
    location_titles: List[str] = ["AREA", "SUBAREA", "EASTING", "NORTHING", "ELEVATION"]
    location_values: List[str]
    location_units: List[str]
    version_titles: List[str] = [
        "CPT DATE",
        "PRINT DATE",
        "PREPARED BY",
        "CHECKED BY",
        "APPROVED BY",
    ]
    version_values: List[str]


class PlotProps(BaseModel):
    plots: PlotContainer
    axes: Axes
    layout: Layout
    info: Info


if __name__ == "__main__":
    print(PlotProps.schema_json(indent=2))



import pandas as pd
import numpy as np
from munch import Munch
import datetime


class P6:
    """
    Class that parses Primavera P6 xer files

    Parameters
    ----------
    path : str
        path to .xer file

    Attributes
    ----------
    path : str
        string of file path
    xer_dict : dict
        Dictionary containing all xer tables and data
    xer_df : dict
        Dictionary of Pandas DataFrames. Keys are as per xer table names.

    Returns
    -------
    object
        P6 object
    """
    def __init__(self, path: str):
        self.path = path
        self.xer_dict = Munch()
        self.xer_df = Munch()

        self._parse_xer()

    def _parse_xer(self):

        with open(self.path, "r") as f:
            xer = f.readlines()

        for line in xer:

            split_line = line.replace("\n", "").split("\t")

            if "%T" in split_line:
                table_name = split_line[-1]
                self.xer_dict[table_name] = Munch()

            elif "%F" in split_line:
                headers = split_line

                for header in headers[1:]:
                    self.xer_dict[table_name][header] = []

            elif "%R" in split_line:
                row = split_line[1:]

                for header, col in zip(headers[1:], row):
                    # inset regex https://stackoverflow.com/questions/12595051/check-if-string-matches-pattern
                    if len(col) > 0:
                        if "date" in header and len(col) > 10:
                            value = datetime.datetime.strptime(col, "%Y-%m-%d %H:%M")
                        else:
                            value = col
                    else:
                        value = np.nan

                    self.xer_dict[table_name][header].append(value)

        for key in self.xer_dict.keys():
            self.xer_df[key.lower()] = pd.DataFrame(self.xer_dict[key])

    def filter_df(self, regex: str = r"([A-E]{1,2}\d[A-D]?)"):
        """
        Filter Task Table by running regex search on the task_name column

        Parameters
        ----------
        regex : str
            a string in regex format. Default value looks for area names at Pulau Tekong Project

        Returns
        -------
        filtered Task DataFrame
        """
        df = self.xer_df.task

        df_filtered = df[
            [
                "task_code",
                "task_name",
                "target_start_date",
                "target_end_date",
                "act_start_date",
                "act_end_date",
            ]
        ]

        df_filtered["start"] = df_filtered["target_start_date"].dt.date
        df_filtered["end"] = df_filtered["target_end_date"].dt.date
        df_filtered["duration"] = np.nan

        if isinstance(regex, str):
            df_filtered["areas"] = df_filtered.task_name.str.findall(regex)

        for index, row in df_filtered.iterrows():
            if isinstance(row["start"], type(pd.NaT)):
                df_filtered.loc[index, "start"] = row["act_start_date"].date()

            if isinstance(row["end"], type(pd.NaT)):
                df_filtered.loc[index, "end"] = row["act_end_date"].date()

        df_filtered["duration"] = (df_filtered["end"] - df_filtered["start"]).dt.days

        return df_filtered

    def to_excel(self, path: str, how: str or list = "all"):
        """
        Method to export xer file to an Excel file

        Parameters
        ----------
        path : str
            String containing file name and/or location.
        how : list of str
            String for list of strings containing the table names to export to Excel. Defaults to 'all'.
            If you want a single sheet to be exported this has to be captured in a list.

        Returns
        -------
        An Excel file with the XER tables per separated per sheet
        """
        if not path.lower().endswith((".xlsx", ".xls")):
            path = path + ".xlsx"

        if how == "all":
            with pd.ExcelWriter(path) as writer:
                for key in self.xer_df:
                    self.xer_df[key].to_excel(writer, sheet_name=key, index=False)

        else:
            with pd.ExcelWriter(path) as writer:
                for key in how:
                    self.xer_df[key].to_excel(writer, sheet_name=key, index=False)

    def __repr__(self):
        return str([key for key in self.xer_dict])



import math
from datetime import datetime

import numpy as np
from pygef import ParseGEF


def standard_gef_plot(gef: ParseGEF) -> dict:
    """

    Parameters
    ----------
    gef : pygef.ParseGef
        parsed gef file by pygef

    Returns
    -------
    plot_props : dict
        Returns a dictionary containing all data and properties for plotting a GEF File
    """

    df = gef.df
    # get max and min depths and round up/down respectively
    ymin = math.floor(df.elevation_with_respect_to_NAP.min())
    ymax = math.ceil(df.elevation_with_respect_to_NAP.max())

    # set the meter tick interval
    y_int = 2

    # make the json object
    plot_props = dict(
        plots=dict(
            y_data=df.elevation_with_respect_to_NAP.values,
            plot_1=dict(
                ax="ax1",
                name="qc",
                type="plot",
                x_data=df.qc.values,
                kwargs=dict(color="black", ls="-", lw=1.5, label="qc"),
            ),
            plot_2=dict(
                ax="ax1_2",
                name="Rf",
                type="plot",
                x_data=df.friction_number.values,
                kwargs=dict(color="red", lw=1, label="Rf"),
            ),
            plot_3=dict(
                ax="ax2",
                name="fs",
                type="plot",
                x_data=df.fs.values,
                kwargs=dict(color="orange", ls="-", lw=1, label="fs"),
            ),
            plot_4=dict(
                ax="ax3",
                name="u2",
                type="plot",
                x_data=df.u2.values,
                kwargs=dict(color="blue", ls="-", lw=1.0, label="u2"),
            ),
        ),
        axes=dict(
            num=3,
            widths=[0.5, 0.25, 0.25],
            kwargs=dict(
                ax1=dict(xlim=[0, 10], xlabel="qt [MPa]", ylabel="Elevation [m+CD]"),
                ax2=dict(xlim=[0, 0.50], xlabel="fs [MPa]"),
                ax3=dict(xlim=[0, 1], xlabel="u2 [MPa]"),
                all=dict(
                    ylim=[ymin, ymax],
                    yticks=[x for x in range(ymin, ymax + y_int, y_int)],
                ),
            ),
            twiny=dict(
                ax1_2=dict(
                    twin="ax1",
                    kwargs=dict(
                        xticks=[x for x in range(0, 6, 1)],
                        xticklabels=["" if x == 0 else str(x) for x in range(0, 6, 1)],
                        xlim=[25, 0],
                        xlabel="Rf [%]",
                    ),
                )
            ),
        ),
        layout=dict(
            papersize="A4",
            margins=[1, 1, 1, 1],
            # logo='./hydronamic.tif'
        ),
        info=dict(
            pointid=[gef.test_id],
            row_spacing=np.linspace(0.6, 0.1, 5),
            general_titles=["CLIENT", "ENGINEER", "AC", "CONTRACTOR", "PROJECT"],
            general_values=[
                "Client A",
                "Consultant A",
                "Consultant B",
                "Boskalis",
                gef.project_id,
            ],
            location_titles=["AREA", "SUBAREA", "EASTING", "NORTHING", "ELEVATION"],
            location_values=["NA", "NA", gef.x, gef.y, gef.zid],
            location_units=["", "", "", "", " m+CD"],
            version_titles=[
                "CPT DATE",
                "PRINT DATE",
                "PREPARED BY",
                "CHECKED BY",
                "APPROVED BY",
            ],
            version_values=[
                gef.file_date.strftime("%d-%b-%Y"),
                datetime.now().strftime("%d-%b-%Y"),
                "MBET",
                "MBET",
                "MBET",
            ],
        ),
    )

    return plot_props



cpt_data = """
SELECT
       *
FROM
     [PulauTekongGintDatgel].[dbo].vw_CPT_DATA_COMB
WHERE PointID = '{pointid}'
ORDER BY Depth
"""

bh_data = """
SELECT
       a.[PointID]
       ,a.[Depth]
       ,[Bottom]
       ,[Elevation] - a.[Depth] AS Elevation_Top
       ,[Elevation] - [Bottom] AS Elevation_Bottom
       ,a.[Depth] - [Bottom] AS Thickness
       ,[Graphic]
      ,[Description]
      ,[Geology_Unit_1]
      ,[Geology_Unit_2]
      ,[Geology_Unit_3]
FROM
     [PulauTekongGintDatgel].[dbo].[STRATA_MAIN] a,
     [PulauTekongGintDatgel].[dbo].[POINT] b
WHERE
      a.PointID = '{pointid}' AND
      b.PointID = '{pointid}' 
ORDER BY
         [Depth]
"""

lab_data = """
SELECT
       *
FROM
     PulauTekongGintDatgel.dbo.vw_BH_ALL
WHERE
      PointID = '{pointid}'
"""

point_data = """
SELECT PointID, Easting, Northing, ElevationStart AS Elevation, DateTime, Area, Box
FROM POINT
WHERE PointID IN {pointids}
"""

point_data_single = """
SELECT PointID, Easting, Northing, ElevationStart AS Elevation, DateTime, Area, Box
FROM POINT
WHERE PointID = '{pointid}'
"""

spt_data = """
SELECT * FROM PulauTekongGintDatgel.dbo.vw_SPT WHERE PointID = '{pointid}' ORDER BY Depth
"""

fvst_data = """
SELECT 
    a.PointID,
    Elevation - Depth AS Elevation,
    Vane_Peak_Uncorrected_Su,
    Vane_Remoulded_Uncorrected_Su
FROM 
    PulauTekongGintDatgel.dbo.IN_SITU_VANE a,
    PulauTekongGintDatgel.dbo.POINT b
WHERE 
    a.PointID = '{pointid}' AND
    b.PointID = '{pointid}'
"""



def cone_resistance(qc, Rf):

    plot_qc = dict(
        ax='ax1',
        name='qc',
        type='plot',
        x_data=qc,
        kwargs=dict(color='black', ls='-', lw=1.5, label='qc')
        )

    plot_Rf = dict(
        ax='ax1_2',
        name='Rf',
        type='plot',
        x_data=Rf,
        kwargs=dict(color='red', lw=1, label='Rf')
        )

    ax1 = ax1=dict(xlim=[0, 10], xlabel='Cone Resistance [MPa]', ylabel='Elevation [m+CD]')
    ax1_2 = dict(
                    twin='ax1',
                    kwargs=dict(
                        xticks=[x for x in range(0, 6, 1)],
                        xticklabels=['' if x == 0 else str(x) for x in range(0, 6, 1)],
                        xlim=[25, 0],
                        xlabel='Rf [%]'
                        )
        )

    return [plot_qc, plot_Rf, ax1, ax1_2]


import math
from datetime import datetime
import textwrap

import numpy as np
import pandas as pd

from .legends_styles import *
from ..sql.gint_sql import *


def gint_plot_cpt_lab_bh(engine, pointid_cpt, pointid_bh):

    # make tuple of the pointids
    pointids = (pointid_cpt, pointid_bh)

    # get data from gINT database. make sure CTP_COMB and BH_COMB views are made
    point = pd.read_sql_query(point_data.format(**{"pointids": str(pointids)}), engine)
    cpt = pd.read_sql_query(cpt_data.format(**{"pointid": str(pointid_cpt)}), engine)
    bh = pd.read_sql_query(bh_data.format(**{"pointid": str(pointid_bh)}), engine)
    lab = (
        pd.read_sql_query(lab_data.format(**{"pointid": str(pointid_bh)}), engine)
        .sort_values("Depth")
        .drop_duplicates()
    )

    # separate the columns for readability of the code
    qt = cpt.Total_Cone_Resistance
    rf = cpt.Friction_Ratio
    fs = cpt.Sleeve_Friction_Resistance
    u0 = cpt.In_Situ_Pore_Pressure
    sigma = cpt.Total_Stress
    sigma_eff = cpt.Effective_Stress
    u2 = cpt.Porewater_Pressure_2
    su = cpt.Undrained_Shear_Strength_1
    pc = cpt.Preconsolidation_Stress_1
    elevation = cpt.Elevation

    lab_y = lab.Elevation
    lab_su_uu = lab.UU_Su
    lab_su_ciu = lab.CIU_Su
    lab_pc = lab.Preconsolidation_Pressure

    bh_x = [0.5] * len(bh)
    bh_height = bh.Thickness
    bh_bottom = bh.Elevation_Top

    # set bh colors
    bh_color = []
    for gu in bh.Geology_Unit_1.values:
        bh_color.append(bh_colormap(gu))

    # get max and min depths and round up/down respectively
    ymin, ymax = math.floor(elevation.min()), math.ceil(elevation.max())

    # set the meter tick interval
    y_int = 2

    # make the json object
    plot_props = dict(
        plots=dict(
            y_data=elevation,
            plot_1=dict(
                ax="ax1",
                name="qc",
                type="plot",
                x_data=qt,
                kwargs=dict(color="black", ls="-", lw=1.5, label="qc"),
            ),
            plot_2=dict(
                ax="ax1_2",
                name="Rf",
                type="plot",
                x_data=rf,
                kwargs=dict(color="red", lw=1, label="Rf"),
            ),
            plot_3=dict(
                ax="ax2",
                name="fs",
                type="plot",
                x_data=fs,
                kwargs=dict(color="orange", ls="-", lw=1, label="fs"),
            ),
            plot_4=dict(
                ax="ax3",
                name="u0",
                type="plot",
                x_data=u0,
                kwargs=dict(color="darkblue", ls="-", lw=1.0, label="u0"),
            ),
            plot_5=dict(
                ax="ax3",
                name="u2",
                type="plot",
                x_data=u2,
                kwargs=dict(color="blue", ls="-", lw=1.0, label="u2"),
            ),
            plot_6=dict(
                ax="ax4",
                name="borehole",
                type="bar",
                x_data=[],
                args=[bh_x, bh_height],
                kwargs=dict(
                    width=0.5,
                    bottom=bh_bottom,
                    color=bh_color,
                    label="BH Log",
                    hatch="...",
                    edgecolor="black",
                ),
            ),
            plot_7=dict(
                ax="ax6",
                name="su",
                type="plot",
                x_data=su,
                kwargs=dict(color="black", label="Su - Nkt"),
            ),
            plot_8=dict(
                ax="ax6",
                name="su_req",
                type="axvline",
                x_data=[],
                args=[23],
                kwargs=dict(color="red", label="Su - Req."),
            ),
            plot_9=dict(
                ax="ax6",
                name="su_lab_uu",
                type="scatter",
                x_data=[],
                args=[lab_su_uu, lab_y],
                kwargs=dict(
                    c="white", s=100, marker="^", edgecolor="black", label="Su - UU"
                ),
            ),
            plot_10=dict(
                ax="ax6",
                name="su_lab_ciu",
                type="scatter",
                x_data=[],
                args=[lab_su_ciu, lab_y],
                kwargs=dict(
                    c="white", s=100, marker="v", edgecolor="black", label="Su - CIU"
                ),
            ),
            plot_11=dict(
                ax="ax5",
                name="u0",
                type="plot",
                x_data=u0,
                kwargs=dict(color="blue", label="u0"),
            ),
            plot_12=dict(
                ax="ax5",
                name="sigma",
                type="plot",
                x_data=sigma,
                kwargs=dict(color="black", label="sigma"),
            ),
            plot_13=dict(
                ax="ax5",
                name="sigma_eff",
                type="plot",
                x_data=sigma_eff,
                kwargs=dict(color="orange", label="sigma_eff"),
            ),
            plot_14=dict(
                ax="ax7",
                name="pc",
                type="plot",
                x_data=pc,
                kwargs=dict(color="orange", label="pc - alpha"),
            ),
            plot_15=dict(
                ax="ax7",
                name="pc_oed",
                type="scatter",
                x_data=[],
                args=[lab_pc, lab_y],
                kwargs=dict(
                    c="white", s=100, marker="^", edgecolor="black", label="Pc - OED"
                ),
            ),
            plot_16=dict(
                ax="ax7",
                name="sigma_eff",
                type="plot",
                x_data=sigma_eff,
                kwargs=dict(color="green", label="sigma_eff"),
            ),
        ),
        axes=dict(
            num=7,
            widths=[0.25, 0.125, 0.125, 0.08, 0.14, 0.14, 0.14],
            kwargs=dict(
                ax1=dict(
                    xlim=[0, 10],
                    xlabel="Cone Resistance [MPa]",
                    ylabel="Elevation [m+CD]",
                ),
                ax2=dict(xlim=[0, 100], xlabel="Friction [kPa]"),
                ax3=dict(xlim=[0, 150], xlabel="Pressure [kPa]"),
                ax4=dict(xlim=[0, 1], xticks=[0, 1], xticklabels=[], xlabel="BH Log"),
                ax5=dict(xlim=[0, 800], xlabel="Stress [kPa]"),
                ax6=dict(xlim=[0, 100], xlabel="SU [kPa]"),
                ax7=dict(xlim=[0, 800], xlabel="Pre-consolidation Stress [kPa]"),
                all=dict(
                    ylim=[ymin, ymax],
                    yticks=[x for x in range(ymin, ymax + y_int, y_int)],
                ),
            ),
            twiny=dict(
                ax1_2=dict(
                    twin="ax1",
                    kwargs=dict(
                        xticks=[x for x in range(0, 6, 1)],
                        xticklabels=["" if x == 0 else str(x) for x in range(0, 6, 1)],
                        xlim=[25, 0],
                        xlabel="Rf [%]",
                    ),
                )
            ),
            custom_legend=dict(
                ax4=bh_legend()
                )
        ),
        layout=dict(
            papersize="A3",
            margins=[1, 1, 1, 1],
            # logo=r'D:\Users\mbet\OneDrive - boskalis.com\9. PEng\a. Pulau Tekong\02. GIS\a. QGIS files\Logo\JV.png'
        ),
        info=dict(
            pointid=[p for p in pointids],
            row_spacing=np.linspace(0.6, 0.1, 5),
            general_titles=["CLIENT", "SO Rep", "PE-GEO", "CONTRACTOR", "PROJECT"],
            general_values=[
                "Housing & Development Board",
                "Surbana Jurong",
                "AECOM",
                "BPJV",
                "Pulau Tekong Polder Development",
            ],
            location_titles=["AREA", "PANEL", "EASTING", "NORTHING", "ELEVATION"],
            location_values=[
                point[x].iloc[0]
                for x in ["Area", "Box", "Easting", "Northing", "Elevation"]
            ],
            location_units=["", "", "", "", " m+CD"],
            version_titles=[
                "CPT DATE",
                "PRINT DATE",
                "PREPARED BY",
                "CHECKED BY",
                "APPROVED BY",
            ],
            version_values=[
                point["DateTime"].iloc[0].strftime("%d-%b-%Y"),
                datetime.now().strftime("%d-%b-%Y"),
                "MBET",
                "ANAM",
                "LBOL",
            ],
        ),
    )

    return plot_props


def standard_gint_cpt_plot(engine, pointid):

    # make tuple of the pointids
    pointids = (pointid, )

    # get data from gINT database. make sure CTP_COMB and BH_COMB views are made
    point = pd.read_sql_query(point_data_single.format(**{"pointid": str(pointid)}), engine)
    cpt = pd.read_sql_query(cpt_data.format(**{"pointid": str(pointid)}), engine)

    # separate the columns for readability of the code
    qt = cpt.Total_Cone_Resistance
    rf = cpt.Friction_Ratio
    fs = cpt.Sleeve_Friction_Resistance
    u0 = cpt.In_Situ_Pore_Pressure
    sigma = cpt.Total_Stress
    sigma_eff = cpt.Effective_Stress
    u2 = cpt.Porewater_Pressure_2
    su = cpt.Undrained_Shear_Strength_1
    pc = cpt.Preconsolidation_Stress_1
    elevation = cpt.Elevation

    # get max and min depths and round up/down respectively
    ymin, ymax = math.floor(elevation.min()), math.ceil(elevation.max())

    # set the meter tick interval
    y_int = 2

    # make the json object
    plot_props = dict(
        plots=dict(
            y_data=elevation,
            plot_1=dict(
                ax="ax1",
                name="qc",
                type="plot",
                x_data=qt,
                kwargs=dict(color="black", ls="-", lw=1.5, label="qc"),
                ),
            plot_2=dict(
                ax="ax1_2",
                name="Rf",
                type="plot",
                x_data=rf,
                kwargs=dict(color="red", lw=1, label="Rf"),
                ),
            plot_3=dict(
                ax="ax2",
                name="fs",
                type="plot",
                x_data=fs,
                kwargs=dict(color="orange", ls="-", lw=1, label="fs"),
                ),
            plot_4=dict(
                ax="ax3",
                name="u0",
                type="plot",
                x_data=u0,
                kwargs=dict(color="darkblue", ls="-", lw=1.0, label="u0"),
                ),
            plot_5=dict(
                ax="ax3",
                name="u2",
                type="plot",
                x_data=u2,
                kwargs=dict(color="blue", ls="-", lw=1.0, label="u2"),
                ),
            ),
        axes=dict(
            num=3,
            widths=[0.50, 0.25, 0.25],
            kwargs=dict(
                ax1=dict(
                    xlim=[0, 10],
                    xlabel="Cone Resistance [MPa]",
                    ylabel="Elevation [m+CD]",
                    ),
                ax2=dict(xlim=[0, 100], xlabel="Friction [kPa]"),
                ax3=dict(xlim=[0, 150], xlabel="Pressure [kPa]"),
                all=dict(
                    ylim=[ymin, ymax],
                    yticks=[x for x in range(ymin, ymax + y_int, y_int)],
                    ),
                ),
            twiny=dict(
                ax1_2=dict(
                    twin="ax1",
                    kwargs=dict(
                        xticks=[x for x in range(0, 6, 1)],
                        xticklabels=["" if x == 0 else str(x) for x in range(0, 6, 1)],
                        xlim=[25, 0],
                        xlabel="Rf [%]",
                        ),
                    )
                ),
            custom_legend=dict(
                ax4=bh_legend()
                )
            ),
        layout=dict(
            papersize="A4",
            margins=[1, 1, 1, 1],
            # logo=r'D:\Users\mbet\OneDrive - boskalis.com\9. PEng\a. Pulau Tekong\02. GIS\a. QGIS files\Logo\JV.png'
            ),
        info=dict(
            pointid=[p for p in pointids],
            row_spacing=np.linspace(0.6, 0.1, 5),
            general_titles=["CLIENT", "SO Rep", "PE-GEO", "CONTRACTOR", "PROJECT"],
            general_values=[
                "Housing & Development Board",
                "Surbana Jurong",
                "AECOM",
                "BPJV",
                "Pulau Tekong Polder Development",
                ],
            location_titles=["AREA", "PANEL", "EASTING", "NORTHING", "ELEVATION"],
            location_values=[
                point[x].iloc[0]
                for x in ["Area", "Box", "Easting", "Northing", "Elevation"]
                ],
            location_units=["", "", "", "", " m+CD"],
            version_titles=[
                "CPT DATE",
                "PRINT DATE",
                "PREPARED BY",
                "CHECKED BY",
                "APPROVED BY",
                ],
            version_values=[
                point["DateTime"].iloc[0].strftime("%d-%b-%Y"),
                datetime.now().strftime("%d-%b-%Y"),
                "MBET",
                "ANAM",
                "LBOL",
                ],
            ),
        )

    return plot_props


def su_gint_cpt_plot(engine, pointid):

    # make tuple of the pointids
    pointids = (pointid, )

    # get data from gINT database. make sure CTP_COMB and BH_COMB views are made
    point = pd.read_sql_query(point_data_single.format(**{"pointid": str(pointid)}), engine)
    cpt = pd.read_sql_query(cpt_data.format(**{"pointid": str(pointid)}), engine)

    # separate the columns for readability of the code
    qt = cpt.Total_Cone_Resistance
    rf = cpt.Friction_Ratio
    fs = cpt.Sleeve_Friction_Resistance
    u0 = cpt.In_Situ_Pore_Pressure
    qnet = cpt.Net_Cone_Resistance
    sigma = cpt.Total_Stress
    sigma_eff = cpt.Effective_Stress
    u2 = cpt.Porewater_Pressure_2
    su = cpt.Undrained_Shear_Strength_1
    pc = cpt.Preconsolidation_Stress_1
    elevation = cpt.Elevation

    # get max and min depths and round up/down respectively
    ymin, ymax = math.floor(elevation.min()), math.ceil(elevation.max())

    # set the meter tick interval
    y_int = 2

    # make the json object
    plot_props = dict(
        plots=dict(
            y_data=elevation,
            plot_1=dict(
                ax="ax1",
                name="qc",
                type="plot",
                x_data=qt,
                kwargs=dict(color="black", ls="-", lw=1.5, label="$q_{t}$"),
                ),
            plot_2=dict(
                ax="ax1_2",
                name="Rf",
                type="plot",
                x_data=rf,
                kwargs=dict(color="red", lw=1, label="$Rf$ [%]"),
                ),
            plot_3=dict(
                ax="ax2",
                name="qnet",
                type="plot",
                x_data=qnet*1000,
                kwargs=dict(color="black", ls="-", lw=1, label="$q_{net}$"),
                ),
            plot_4=dict(
                ax="ax3",
                name="u0",
                type="plot",
                x_data=u0,
                kwargs=dict(color="darkblue", ls="-", lw=1.0, label="$u_{0}$"),
                ),
            plot_5=dict(
                ax="ax3",
                name="u2",
                type="plot",
                x_data=u2,
                kwargs=dict(color="blue", ls="-", lw=1.0, label="$u_{2}$"),
                ),
            plot_6=dict(
                ax="ax3",
                name="sigma",
                type="plot",
                x_data=sigma,
                kwargs=dict(color="black", ls="--", lw=1.0, label="$\sigma_{v}$"),
                ),
            plot_7=dict(
                ax="ax3",
                name="sigma",
                type="plot",
                x_data=sigma_eff,
                kwargs=dict(color="black", ls="-", lw=1.0, label="$\sigma'_{v}$"),
                ),
            plot_10=dict(
                ax="ax4",
                name="su",
                type="plot",
                x_data=su,
                kwargs=dict(color="orange", ls="-", lw=1.0, label="$su$"),
                ),
            plot_11=dict(
                ax="ax4",
                name="su_req",
                type="axvline",
                x_data=[],
                args = [23],
                kwargs=dict(color="red", ls="-", lw=1.0, label="Req."),
                ),
            plot_8=dict(
                ax="ax4",
                name="sigma_eff",
                type="plot",
                x_data=sigma_eff * 0.23,
                kwargs=dict(color="black", ls="--", lw=1.0, label="$.23 * \sigma'_{v}$"),
                ),
            plot_9=dict(
                ax="ax4",
                name="sigma_eff",
                type="plot",
                x_data=sigma_eff * 0.28,
                kwargs=dict(color="black", ls="-.", lw=1.0, label="$.28 * \sigma'_{v}$"),
                ),
            ),
        axes=dict(
            num=4,
            widths=[0.25, 0.25, 0.25, 0.25],
            kwargs=dict(
                ax1=dict(
                    xlim=[0, 10],
                    xlabel="Cone Resistance [MPa]",
                    ylabel="Elevation [m+CD]",
                    ),
                ax2=dict(xlim=[0, 1000], xlabel="Nett. Cone Resistance [kPa]"),
                ax3=dict(xlim=[0, 500], xlabel="In-situ Stress [kPa]"),
                ax4=dict(xlim=[0, 50], xlabel="SU [kPa]"),
                all=dict(
                    ylim=[ymin, ymax],
                    yticks=[x for x in range(ymin, ymax + y_int, y_int)],
                    ),
                ),
            twiny=dict(
                ax1_2=dict(
                    twin="ax1",
                    kwargs=dict(
                        xticks=[x for x in range(0, 6, 1)],
                        xticklabels=["" if x == 0 else str(x) for x in range(0, 6, 1)],
                        xlim=[25, 0],
                        xlabel="",
                        ),
                    )
                ),
            ),
        layout=dict(
            papersize="A4",
            margins=[1, 1, 1, 1],
            # logo=r'D:\Users\mbet\OneDrive - boskalis.com\9. PEng\a. Pulau Tekong\02. GIS\a. QGIS files\Logo\JV.png'
            ),
        info=dict(
            pointid=[p for p in pointids],
            row_spacing=np.linspace(0.6, 0.1, 5),
            general_titles=["CLIENT", "SO Rep", "PE-GEO", "CONTRACTOR", "PROJECT"],
            general_values=[
                "Housing & Development Board",
                "Surbana Jurong",
                "AECOM",
                "BPJV",
                "Pulau Tekong Polder Development",
                ],
            location_titles=["AREA", "PANEL", "EASTING", "NORTHING", "ELEVATION"],
            location_values=[
                point[x].iloc[0]
                for x in ["Area", "Box", "Easting", "Northing", "Elevation"]
                ],
            location_units=["", "", "", "", " m+CD"],
            version_titles=[
                "CPT DATE",
                "PRINT DATE",
                "PREPARED BY",
                "CHECKED BY",
                "APPROVED BY",
                ],
            version_values=[
                point["DateTime"].iloc[0].strftime("%d-%b-%Y"),
                datetime.now().strftime("%d-%b-%Y"),
                "MBET",
                "ANAM",
                "LBOL",
                ],
            ),
        )

    return plot_props


def standard_gint_bh_plot(engine, pointid):
    # make tuple of the pointids
    pointids = (pointid, )

    # get data from gINT database. make sure CTP_COMB and BH_COMB views are made
    point = pd.read_sql_query(point_data_single.format(**{"pointid": str(pointid)}), engine)
    bh = pd.read_sql_query(bh_data.format(**{"pointid": str(pointid)}), engine)
    spt = pd.read_sql_query(spt_data.format(**{"pointid": str(pointid)}), engine)
    vst = pd.read_sql_query(fvst_data.format(**{"pointid": str(pointid)}), engine)
    lab = (
        pd.read_sql_query(lab_data.format(**{"pointid": str(pointid)}), engine)
            .sort_values("Depth")
            .drop_duplicates()
    )

    # separate the columns for readability of the code
    elevation = bh.Elevation_Top

    lab_y = lab.Elevation
    lab_su_uu = lab.UU_Su
    lab_su_ciu = lab.CIU_Su

    bh_x = [0.5] * len(bh)
    bh_height = bh.Thickness
    bh_bottom = bh.Elevation_Top

    # set bh colors
    bh_color = []
    for gu in bh.Geology_Unit_1.values:
        bh_color.append(bh_colormap(gu))

    # get max and min depths and round up/down respectively
    ymin, ymax = math.floor(elevation.min())-1, math.ceil(elevation.max())

    # set the meter tick interval
    y_int = 2

    # makersize lab and vst
    s=50

    # make the json object
    plot_props = dict(
        plots=dict(
            y_data=elevation,
            plot_1=dict(
                ax="ax1",
                name="borehole",
                type="bar",
                x_data=[],
                args=[bh_x, bh_height],
                kwargs=dict(
                    width=0.5,
                    bottom=bh_bottom,
                    color=bh_color,
                    label="BH Log",
                    hatch="...",
                    edgecolor="black",
                    ),
                ),
            plot_2=dict(
                ax="ax2",
                name="Description",
                type="hlines",
                x_data=[],
                args=[elevation, 0, 1],
                kwargs=dict(color='black', lw=0.8),
                ),
            plot_3=dict(
                ax="ax3",
                name="borehole",
                type="plot",
                x_data=[],
                args=[spt.N_Value, spt.Elevation],
                kwargs=dict(color='black', marker='o', ms=10, mec='black', mfc='red', label='SPT-N Value'),
                ),
            plot_4=dict(
                ax="ax4",
                name="uu",
                type="scatter",
                x_data=[],
                args=[lab_su_uu, lab_y],
                kwargs=dict(marker='^', s=s, edgecolors='black', c='white', label='UU'),
                ),
            plot_5=dict(
                ax="ax4",
                name="ciu",
                type="scatter",
                x_data=[],
                args=[lab_su_ciu, lab_y],
                kwargs=dict(marker='v', s=s, edgecolors='black', c='white', label='CIU'),
                ),
            plot_6=dict(
                ax="ax4",
                name="vst",
                type="scatter",
                x_data=[],
                args=[vst.Vane_Peak_Uncorrected_Su, vst.Elevation],
                kwargs=dict(marker='<', s=s, edgecolors='black', c='white', label='F-VST'),
                )
            ),
        axes=dict(
            num=4,
            widths=[0.15, 0.30, 0.30, 0.25],
            kwargs=dict(
                ax1=dict(xlim=[0, 1], xticks=[0, 1], xticklabels=[], xlabel="BH Log", ylabel='Elevation [m+CD]'),
                ax2=dict(xlim=[0, 1], xticks=[0, 1], yticklabels=[], xticklabels=[], xlabel="Description"),
                ax3=dict(xlim=[0, 100], xlabel="SPT [-]"),
                ax4=dict(xlim=[0, 100], xlabel="Undrained Shear Stength [kPa]"),
                all=dict(
                    ylim=[ymin, ymax],
                    yticks=[x for x in range(ymin, ymax + y_int, y_int)],
                    ),
                ),
            custom_legend=dict(
                ax1=bh_legend()
                ),
            no_grid=['ax2']
            ),
        layout=dict(
            papersize="A4",
            margins=[1, 1, 1, 1],
            # logo=r'D:\Users\mbet\OneDrive - boskalis.com\9. PEng\a. Pulau Tekong\02. GIS\a. QGIS files\Logo\JV.png'
            ),
        info=dict(
            pointid=[p for p in pointids],
            row_spacing=np.linspace(0.6, 0.1, 5),
            general_titles=["CLIENT", "SO Rep", "PE-GEO", "CONTRACTOR", "PROJECT"],
            general_values=[
                "Housing & Development Board",
                "Surbana Jurong",
                "AECOM",
                "BPJV",
                "Pulau Tekong Polder Development",
                ],
            location_titles=["AREA", "PANEL", "EASTING", "NORTHING", "ELEVATION"],
            location_values=[
                point[x].iloc[0]
                for x in ["Area", "Box", "Easting", "Northing", "Elevation"]
                ],
            location_units=["", "", "", "", " m+CD"],
            version_titles=[
                "CPT DATE",
                "PRINT DATE",
                "PREPARED BY",
                "CHECKED BY",
                "APPROVED BY",
                ],
            version_values=[
                point["DateTime"].iloc[0].strftime("%d-%b-%Y"),
                datetime.now().strftime("%d-%b-%Y"),
                "MBET",
                "ANAM",
                "LBOL",
                ],
            ),
        )

    for index, row in bh.iterrows():
        plot_name = f'plot_text_{index}'
        text = '\n'.join(textwrap.wrap(row.Description, width=40))
        props = dict(
            ax='ax2',
            type='annotate',
            x_data=[],
            args=[text, (0.05, row.Elevation_Top)],
            kwargs=dict(va='top', ha='left', fontsize=7, wrap=True)
        )

        plot_props['plots'][plot_name] = props



    return plot_props


import re

from munch import Munch
import pandas as pd


class AGS:
    """Class to parse AGS files.

    Parameters
    ----------
    path : str
        path to AGS file
    ags_version : int
        specify AGS version. Either 3 or 4. Defaults to 3, as 4 still needs to be implemented.

    Attributes
    ----------
    ags_version : int
        version number of the AGS file
    ags_raw : list
        list of strings containing the raw AGS ASCII data
    ags : dict
        dictionary containing all tables and data
    df : Munch dict
        dictionary of dataframe. Each AGS table is represented in a dataframe containing all the data
    """

    def __init__(self, path, ags_version=3):

        self.ags_version = ags_version

        with open(path, 'r') as f:
            self.ags_raw = f.readlines()

        self.ags = dict()
        self.df = Munch()

        if self.ags_version == 3:
            self._parse_ags_3()
            self._make_df()
        elif self.ags_version == 4:
            self._parse_ags_4()
            self._make_df()
        else:
            raise AttributeError('This AGS version is unknown. Please only use AGS 3 or AGS 4.')

    def _parse_ags_3(self):
        """
        private method to parse AGS3 data

        """
        for i, line in enumerate(self.ags_raw):
            if '**' in line:
                group = re.search(r'[A-Z]{4}', line).group(0)
                group_line = i
                if '?' in line:
                    user_def = True
                else:
                    user_def = False

                self.ags[group] = dict()

            elif '*' in line:
                headers = re.findall(r'[A-Z]{4}_[A-Z,0-9]{2,4}\d?', line)
                for header in headers:
                    self.ags[group][header] = []

                start_block = i + 1

            elif '<UNITS>' in line:
                pass

            else:
                split_line = line.split('","')
                len_block = len(split_line)
                len_header = len(self.ags[group].keys())
                headers = list(self.ags[group].keys())

                if len_block == len_header:
                    if '"<CONT>' == split_line[0]:
                        for header, col in zip(headers[1:], split_line[1:]):
                            self.ags[group][header][-1] = self.ags[group][header][-1] + (
                                col.replace('\n', '').replace('"', ''))
                    else:
                        for header, col in zip(headers, split_line):
                            self.ags[group][header].append(col.replace('\n', '').replace('"', ''))

    def _parse_ags_4(self):
        """
        private method to parse AGS4 data
        """
        for i, line in enumerate(self.ags_raw):
            if 'GROUP' in line:
                group = re.search(r'"[A-Z]{4}"', line).group(0).replace('"','')
                group_line = i

                self.ags[group] = dict()

            elif 'HEADING' in line:
                headers = re.findall(r'[A-Z]{4}_[A-Z,0-9]{2,4}\d?', line)
                for header in headers:
                    self.ags[group][header] = []

                start_block = i + 1

            elif 'UNIT' in line:
                # TODO add parser for unit and convert columns to other unit if required
                pass

            elif 'DATA' in line:
                split_line = line.split('","')[1:]
                len_block = len(split_line)
                len_header = len(self.ags[group].keys())
                headers = list(self.ags[group].keys())

                for header, col in zip(headers, split_line):
                    self.ags[group][header].append(col.replace('\n', '').replace('"', ''))

            else:
                pass

    def _make_df(self, to_numeric=True):
        """Private method to make DataFrames from the AGS dict data

        Parameters
        ----------
        to_numeric : bool
            True if values should be converted to numeric values. Default is true
        """
        for key in self.ags.keys():
            self.df[key] = pd.DataFrame(self.ags[key])

            if to_numeric:
                self.df[key] = self.df[key].apply(pd.to_numeric, errors='ignore')

    def _gint_mapper(self, mapper):
        """function to rename ags columns in the DataFrames

        Parameters
        ----------
        mapper : dict
            mapper on how the rename the extracted DataFrames. Used for uploading to databases.

        Returns
        -------
        renamed dataframe
        """
        renamed_df = Munch()

        for group in mapper.group.unique():
            try:
                temp = mapper[mapper.group == group][['ags_name', 'gint_name']]
                headers = temp.ags_name.to_list()
                d = {x[0]: x[1] for x in temp.values}

                temp_df = self.df[group][headers].rename(columns=d)

                for col in temp_df.columns:
                    if 'date' in col.lower():
                        temp_df[col] = pd.to_datetime(temp_df[col], format=r'%d/%m/%Y', errors='coerce')

                renamed_df[group] = temp_df
            except KeyError:
                renamed_df[group] = pd.DataFrame()

        return renamed_df

    @staticmethod
    def upload_to_gint(engine, df_dict, sequence, gint_proj_id=3):
        # TODO Make uploader to gint using merge statement and a reflected gINT database
        pass


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['boka_tools',
 'boka_tools.planning',
 'boka_tools.soil_investigation',
 'boka_tools.soil_investigation.boiler_plates',
 'boka_tools.soil_investigation.json',
 'boka_tools.soil_investigation.sql',
 'boka_tools.utils']

package_data = \
{'': ['*']}

install_requires = \
['SQLAlchemy==1.4.27', 'pandas>=1.3.4,<2.0.0']

setup_kwargs = {
    'name': 'boka-tools',
    'version': '0.3.0',
    'description': 'package for general automation work at Boskalis',
    'long_description': None,
    'author': 'Maarten Betman',
    'author_email': 'maarten.betman@boskalis.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



"""function to print gantt charts

Code by: Maarten Betman
DTED Hydronamic, Boskalis
maarten.betman@boskalis.com
date 05-Jun-2020

"""
# Standard libs
import datetime
from collections import OrderedDict
import math
from calendar import monthrange
import warnings

# Other libs
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
import matplotlib.patches as patches
import matplotlib.image as image
from pandas.plotting import register_matplotlib_converters

register_matplotlib_converters()

# paper sizes in inches, landscape
PAPER_SIZES = dict(
    A0=[46.8, 33.1], A1=[33.1, 23.4], A2=[23.4, 16.5], A3=[16.5, 11.7], A4=[11.7, 8.3],
)


def gantt(
    df,
    datemin,
    datemax,
    date_cols_prefix=None,
    date_cols_sep="_",
    date_cols_suffix=None,
    date_bars_widths=None,
    date_bars_color_categories="",
    date_bars_color_type="planning",
    date_bars_colors=None,
    col_widths=None,
    footer_info=None,
    plt_title="Planning",
    papersize="A1",
    object_lines=16,
    max_objects=5,
    cols_table=None,
):
    """
    function to plot a gantt chart of a planning DataFrame

    Parameters
    ----------
    df : Pandas DataFrame
        Dataframe with activities to plot. Most comply with XER format in terms of columns names and defined pre- and
        suffixes for dates
    datemin : Datetime date
        Start date of plot
    datemax : Datetime date
        End date of plot
    date_cols_prefix : list
        list with prefix for column names, default ['Start', 'End']
    date_cols_sep : str
        separator between prefix and suffix defining the date columns, default "_"
    date_cols_suffix : list
        list with suffic for column names, default ['Start', 'End']
    date_bars_widths : list
        specify the height of the gantt bars. default [3, 1.5]
    date_bars_color_categories : str
        Define how gantt function should distinct between categories. Either column name or empty string to be
        passed. If empty string is passed distinction is made between the date categories (eg. planned / actual).
        default ''
    date_bars_color_type : str
        default 'planning'
    date_bars_colors : list
        Colors to use for gantt bars. Length should correspond with categories. default ['#65f794','blue']
    col_widths : list
    footer_info : dict
    plt_title : str
        title on top of gantt chart, default Planning
    papersize : str
        iso paper size format, default 'A1'
    object_lines : int
        number of lines be obs object, default 16
    max_objects : int
        number of obs object to plot, default 5
    cols_table : list
        df columns names for table, default ['Box', 'Task', 'Planned_Start', 'Planned_End', 'Actual_Start', 'Actual_End']

    Returns
    -------
    figure
        returns a figure containing a gantt chart

    Raises
    ------
    ValueError
        if lists are not corresponding in length

    """
    # cm to inches
    if date_cols_prefix is None:
        date_cols_prefix = ["Planned", "Actual"]
    if date_cols_suffix is None:
        date_cols_suffix = ["Start", "End"]
    if date_bars_widths is None:
        date_bars_widths = [3, 1.5]
    if date_bars_colors is None:
        date_bars_colors = [
            "#65f794",
            "blue",
        ]
    if cols_table is None:
        cols_table = [
            "Box",
            "Planned_Start",
            "Task",
            "Planned_End",
            "Actual_Start",
            "Actual_End",
        ]

    cm_to_inch = 0.393700787

    # font sizes
    font_size_table = 9
    font_size_title = 24
    font_size_footer = 12

    # footer info
    if footer_info is None:
        footer_info = dict(
            legend_title="Legend",
            page=1,
            total_pages=1,
            prepared_by="Mr. X",
            checked_by="Mr. Y",
            aproved_by="Mrs. Z",
            date=datetime.datetime.now().strftime("%d-%b-%Y"),
        )

    # Margins around planning in inch
    MARGINS = dict(
        top=(cm_to_inch * 4) / PAPER_SIZES[papersize][1],
        left=(cm_to_inch * 1) / PAPER_SIZES[papersize][0],
        right=(cm_to_inch * 1) / PAPER_SIZES[papersize][0],
        bottom=(cm_to_inch * 5) / PAPER_SIZES[papersize][1],
    )

    # Get bbox dimensions for planning axis
    width_table = 0.3
    rect = [
        width_table + MARGINS["right"],
        MARGINS["bottom"],
        (1 - MARGINS["left"] - MARGINS["right"] - width_table),
        (1 - MARGINS["top"] - MARGINS["bottom"]),
    ]  # xll, yll, width, height

    rect_table = [
        MARGINS["right"],
        MARGINS["bottom"],
        width_table,
        (1 - MARGINS["top"] - MARGINS["bottom"]),
    ]

    rect_header = [
        MARGINS["right"],
        (1 - MARGINS["top"] + (cm_to_inch * 1.5) / PAPER_SIZES[papersize][1]),
        (1 - MARGINS["right"] - MARGINS["left"]),
        (cm_to_inch * 2.0) / PAPER_SIZES[papersize][1],
    ]

    rect_footer = [
        MARGINS["right"],
        (cm_to_inch * 1) / PAPER_SIZES[papersize][1],
        (1 - MARGINS["right"] - MARGINS["left"]),
        (cm_to_inch * 3.75) / PAPER_SIZES[papersize][1],
    ]

    rect_img = [
        MARGINS["right"],
        (1 - MARGINS["top"] + (cm_to_inch * 1.7) / PAPER_SIZES[papersize][1]),
        (cm_to_inch * 4.0) / PAPER_SIZES[papersize][0],
        (cm_to_inch * 1.75) / PAPER_SIZES[papersize][1],
    ]

    # place logo in header
    # im = image.imread("./resources/JV.png") TODO Add logo support for top of gantt

    # make ax and figure
    fig = plt.figure(figsize=PAPER_SIZES[papersize], edgecolor="black", linewidth=3.0)
    ax = plt.axes(rect)
    ax_legend = plt.axes(rect_footer)
    ax_header = plt.axes(rect_header)
    # im_ax = fig.add_axes(rect_img, anchor="NE", zorder=10)
    ax_table = plt.axes(rect_table)

    x = 0
    for w in [1 / 3] * 3:
        ax_legend.add_patch(
            patches.Rectangle(
                (x, 0), w, 1, edgecolor="black", facecolor="none", linewidth=1.5
            )
        )
        x = x + w

    ax_table.axis("off")
    # im_ax.axis("off")
    ax_legend.set_xticks([])
    ax_legend.set_yticks([])
    ax_header.set_xticks([])
    ax_header.set_yticks([])

    # im_ax.imshow(im)

    # set styling dimension of chart
    barwidth_planned = 3
    barwidth_actual = barwidth_planned / 2
    activitywidth = 5
    whitespace_around_bars = 1

    # calc max number of lines
    max_lines = object_lines * max_objects

    # Date ranges for month and week headers
    date_range = np.arange(datemin, datemax, np.timedelta64(1, "M")).astype(
        datetime.date
    )
    date_range_total_days = sum([monthrange(x.year, x.month)[1] for x in date_range])
    week_nums = _find_weeks(
        datemin.astype(datetime.datetime), datemax.astype(datetime.date)
    )
    wk = _week_width(datemin.astype(datetime.date), datemax.astype(datetime.date))

    # add lines to df if maxlines is not satified
    missing_rows = max_lines - df.shape[0]
    for i in range(missing_rows):
        df = df.append(pd.Series(), ignore_index=True)

    # y-axis formatting
    ax.set_yticks(np.arange(0, 3.5 + (max_lines + 1) * activitywidth, activitywidth))
    ax.set_yticklabels([])
    ax.set_ylim(max_lines * activitywidth, 0)

    # x-axis tick locators
    months = mdates.MonthLocator()  # every month
    weeks = mdates.WeekdayLocator()
    date_fmt = mdates.DateFormatter("%Y-%b")

    # x-axis ticks
    ax.xaxis.set_major_locator(months)
    ax.xaxis.set_major_formatter(date_fmt)
    ax.xaxis.set_minor_locator(weeks)

    ax.set_xlim(datemin, datemax)
    ax.xaxis.tick_top()
    ax.set_xticklabels([])

    # error checking
    if not len(date_cols_suffix) == 2:
        raise ValueError("Only a start and end suffix are allowed")

    if len(df) > max_lines:
        df = df.iloc[:80, :]
        warnings.warn("DataFrame truncated because it was more than 80 lines")

    # draw bars
    pos = 0
    row_colors = ["#e8e8e8", "white"] * 10
    rowColours = []

    if date_bars_color_type == "pandas_col":
        date_bars_color_category_list = list(df[date_bars_color_categories].unique())
    elif date_bars_color_type == "list":
        date_bars_color_category_list = date_bars_color_categories
    elif date_bars_color_type is None:
        pass
    else:
        pass

    # Plot bars
    for index, row in df.iterrows():
        for i, prefix in enumerate(date_cols_prefix):
            if date_bars_color_type == "pandas_col":
                cat = row[date_bars_color_categories]
                i_c = date_bars_color_category_list.index(cat)
                label = cat
            else:
                label = prefix
                i_c = i

            start = row[date_cols_sep.join([prefix, date_cols_suffix[0]])]
            end = row[date_cols_sep.join([prefix, date_cols_suffix[1]])]
            middle = activitywidth / 2 + pos * activitywidth

            if not isinstance(start, type(pd.NaT)):
                ax.broken_barh(
                    [(start, end - start)],
                    (middle - date_bars_widths[i] / 2, date_bars_widths[i]),
                    facecolor=date_bars_colors[i_c],
                    edgecolor="black",
                    label=label,
                )

        rowColours.append(row_colors[math.floor(pos / object_lines)])

        pos += 1

    header_height = 0.0125 / 0.75

    the_table = plt.table(
        cellText=df[cols_table]
        .astype(str)
        .replace("NaT", "")
        .replace("nan", "")
        .values,
        colWidths=col_widths,  # TODO Make something to look at string length in cell
        cellLoc="center",
        cellColours=[[x] * len(cols_table) for x in rowColours],
        bbox=(0.0, 0.0, 1, 1),
    )

    if col_widths is None:
        w = the_table.get_celld()[0, 0].get_width()

        for i in range(len(cols_table)):
            the_table.add_cell(
                -1, i, height=header_height, width=w, text=cols_table[i], loc="center"
            )

    else:
        for i, w in enumerate(col_widths):
            the_table.add_cell(
                -1, i, height=header_height, width=w, text=cols_table[i], loc="center"
            )

    # Grey bands
    for i in range(max_objects):
        if (i % 2) == 0:
            ax.axhspan(
                i * max_lines,
                (i + 1) * max_lines,
                facecolor="lightgrey",
                alpha=0.5,
                zorder=-100,
            )

    # Date line
    ax.axvline(datetime.datetime.now(), color="red", label="Today")

    ax.margins(y=(1 - barwidth_planned) / 2 / pos)
    ax.grid(linestyle="--", linewidth=1)

    # set title
    fig.suptitle(plt_title, fontsize=font_size_title)

    # Create legend and remove repeated label due to broken bar loop
    handles, labels = ax.get_legend_handles_labels()
    by_label = OrderedDict(zip(labels, handles))
    ax_legend.legend(
        by_label.values(),
        by_label.keys(),
        loc="upper left",
        bbox_to_anchor=(0.005, 0.8),
        frameon=False,
        fontsize=font_size_table,
        ncol=4,
    )

    # add months at top of gantt chart
    table_dates = ax.table(
        cellText=[[x.strftime("%Y-%b") for x in date_range]],
        colWidths=[
            monthrange(x.year, x.month)[1] / date_range_total_days for x in date_range
        ],
        cellLoc="center",
        bbox=(0.0, 1.0125, 1.0, 0.0125),
    )

    # add weeks at top of gantt chart
    table_weeks = ax.table(
        cellText=[week_nums],
        colWidths=[
            (j - i).days / date_range_total_days for i, j in zip(wk[:-1], wk[1:])
        ],
        cellLoc="center",
        bbox=(0.0, 1.0, 1.0, 0.0125),
    )

    # redraw to change table fonts
    for t in [
        table_dates,
        table_weeks,
        the_table,
    ]:
        t.auto_set_font_size(False)
        t.set_fontsize(font_size_table)

    # insert text in footer
    ax_legend.text(
        0.005,
        0.9,
        footer_info["legend_title"],
        fontsize=font_size_footer,
        fontweight="bold",
        verticalalignment="top",
        horizontalalignment="left",
    )

    ax_legend.text(
        0.5,
        0.1,
        f"[page {footer_info['page']} of {footer_info['total_pages']}]",
        fontsize=font_size_footer,
        verticalalignment="center",
        horizontalalignment="center",
    )

    for i, key in enumerate(list(footer_info.keys())[-4:]):
        y_start = 0.65
        x_start = 0.9
        spacing_lines = 0.15
        ax_legend.text(
            x_start,
            y_start - (i * spacing_lines),
            key,
            fontsize=font_size_footer,
            verticalalignment="top",
            horizontalalignment="left",
        )
        ax_legend.text(
            x_start + 0.05,
            y_start - (i * spacing_lines),
            f":",
            fontsize=font_size_footer,
            verticalalignment="top",
            horizontalalignment="left",
        )
        ax_legend.text(
            x_start + 0.06,
            y_start - (i * spacing_lines),
            footer_info[key],
            fontsize=font_size_footer,
            verticalalignment="top",
            horizontalalignment="left",
        )

    fig.canvas.draw()
    plt.close()

    return fig


def _week_width(start, end):
    l = [start]
    for i in range((end - start).days + 1):
        d = start + datetime.timedelta(days=i)
        if d.weekday() == 1:
            l.append(d)
    l.append(end)
    return l


def _find_weeks(start, end):
    l = []
    for i in range((end - start).days + 1):
        d = (start + datetime.timedelta(days=i + 1)).isocalendar()  # e.g. (2011, 52)
        yearweek = "{}{:02}".format(*d)  # e.g. "201152"
        l.append(yearweek)
    return [x[-2:] for x in sorted(set(l))]



from matplotlib.colors import ListedColormap, BoundaryNorm
from matplotlib.patches import Patch


def ic_colormap():
    """

    Returns
    -------

    """
    ic_colors = ["#d6a000", "#d2d600", "#02e595", "#005f9b", "#005f9b", "#7f5901"]
    ic_values = [0, 1.31, 2.05, 2.60, 2.95, 3.60]
    # ic_labels = ["Gravel", "Sands", "Sand mixtures", "Silt mixtures", "Clays", "Peats"]

    cmap = ListedColormap(ic_colors)
    norm = BoundaryNorm(ic_values, cmap.N, clip=True)

    return cmap, norm


def ic_legend():
    """
    Function to make a legend for Ic plots.
    Ic = Soil Behaviour type index

    Returns
    -------
    legend : list
        Returns a list with legend patches

    """
    ic_colors = ["#d6a000", "#d2d600", "#02e595", "#005f9b", "#005f9b", "#7f5901"]
    ic_values = [0, 1.31, 2.05, 2.60, 2.95, 3.60]
    ic_labels = ["Gravel", "Sands", "Sand mixtures", "Silt mixtures", "Clays", "Peats"]

    legend = []

    for c, s, l in zip(ic_colors, ic_values, ic_labels):
        entry = Patch(facecolor=c, edgecolor="black", label=l)
        legend.append(entry)

    return legend


def bh_colormap(su_in_log: str):
    """
    Function that returns the color code for the corresponding geological formation

    Parameters
    ----------
    su_in_log : str
        GEOL_LEG entry in the borehole log

    Returns
    -------
    color : str
        Returns string with correct color code
    """
    su_in_log_stripped = su_in_log.strip()

    colors = [
        "#E6E67D",
        "#99CCFF",
        "#CCFF99",
        "#FFFF99",
        "#9999FF",
        "#C0C0C0",
        "#C0C0C0",
        "#C0C0C0",
        "#C0C0C0",
        "#C0C0C0",
        "#C0C0C0",
        "#C0C0C0",
        "#878686",
        "#878686",
        "#878686",
        "#878686",
        "#878686",
        "#878686",
    ]


    su = [
        "FILL",
        "M",
        "E",
        "F1",
        "F2",
        "S",
        "S(I)",
        "S(II)",
        "S(III)",
        "S(IV)",
        "S(V)",
        "S(VI)",
        "G(V)",
        "G(IV)",
        "G(VI)",
        "G(III)",
        "G(II)",
        "G(I)",
    ]
    labels = [
        "Sand (fill)",
        "Marine Clay",
        "Estuarine Clay",
        "Fluvial Sand",
        "Fluvial Clay",
        "Residual Soil",
        "Granite"
    ]

    if su_in_log_stripped not in su:
        raise ValueError(f"Soil Unit not known - {su_in_log_stripped}")

    return colors[su.index(su_in_log_stripped)]


def bh_legend():
    """
    Function to make the borehole legend

    Returns
    -------
    legend : list
        returns a list with legend patches
    """
    colors = [
        "#E6E67D",
        "#99CCFF",
        "#CCFF99",
        "#FFFF99",
        "#9999FF",
        "#C0C0C0",
        "#878686",
    ]

    su = ["FILL", "M", "E", "F1", "F2", "S"]

    labels = [
        "Sand (fill)",
        "Marine Clay",
        "Estuarine Clay",
        "Fluvial Sand",
        "Fluvial Clay",
        "Residual Soil",
        "Granite"
    ]

    legend = []

    for c, s, l in zip(colors, su, labels):
        entry = Patch(facecolor=c, edgecolor="black", label=f"{s} - {l}")
        legend.append(entry)

    return legend



"""
Methods to make sql engines and query spatial data from MS SQL Server and return it into a pandas DataFrame.
"""
import platform

import pandas as pd
import urllib.parse
import sqlalchemy as sa

try:
    import geopandas as gpd
    from shapely.wkt import loads
except ImportError:
    pass


def server_connect(
        server: str, database: str, trusted_connection: bool = False, user: str = "api", password: str = "api",
) -> object:
    """
    creates and server engine that can be used to query data from SQL server using Pandas

    Parameters
    -----------
    server : str
        Server location
    database : str
        Database name
    trusted_connection : bool
        Designates if the connection is Trusted or not,
        When set to False and correct username and password must be entered, Boolean
    user : str
        Optional, username used to access the server if TrustedConnection is False
    password : str
        Optional, password to entered username to access the server

    Returns
    -------
    object
        Returns a SQLAlchemy engine
    """
    if platform.system().lower() == 'windows':
        driver = 'SQL Server'
    else:
        driver = 'ODBC Driver 17 for SQL Server'

    # connection string for database
    if trusted_connection:
        params = urllib.parse.quote_plus(
            "Driver={"+driver+"};Server="
            + server
            + ";Database="
            + database
            + ";Trusted_Connection=yes"
        )
    else:
        params = urllib.parse.quote_plus(
            "Driver={"+driver+"};Server="
            + server
            + ";Database="
            + database
            + ";UID="
            + user
            + ";PWD="
            + password
        )

    engine = sa.create_engine("mssql+pyodbc:///?odbc_connect=%s" % params)

    return engine


def mssql_geom(sql: str, engine: object, epsg: int, to_latlon: bool=False) -> object:
    """
    Pass Microsoft SQL statement and return a Geopandas dataframe
    Geometry columns should be named "geometry" in sql statement.

    Parameters
    ----------
    sql : str
        T-SQL statement with a Spatial geometry column passed as 'geometry'
        in Well Known Text (wkt) via the T-SQL function > [GEOM].STAsText()
    engine : object
        SQL Alchemy engine should be passed into function. Assuming engine
        is already made under the variable 'engine' it is passed by default
    epsg : int
        the epsg code of the geometery you are importing
    to_latlon : bool
        whether or not to convert the geometry to lat lon coordinates in epsg:4326

    Returns
    -------
    object
        Returns a GeoPandas DataFrame

    Raises
    ------
    ValueError
        is STAsText T-SQL Syntac is not detected in sql string

    """

    if 'STAsText' not in sql:
        raise ValueError('GEOM column not passed as WKT by means of [GEOM].STAsText()')

    df = pd.read_sql_query(sql, engine)
    df['geometry'] = df['geometry'].apply(loads)
    gdf = gpd.GeoDataFrame(df, geometry='geometry', crs={'init': f'epsg:{epsg}'})

    if to_latlon:
        gdf = gdf.to_crs({'init': 'epsg:4326'})

    return gdf
